
/*
   ITISWW
   Prototipo di dispositivo di early warning per
   "MONITORAGGIO DEI CORSI D'ACQUA PER LA SICUREZZA DEL TERRITORIO
   E PER SALVARE VITE UMANE"

   Progetto realizzato dal Polo Tecnologico Imperiese nell'ambito del
   "PREMIO ROTARY CLUB DEL PONENTE LIGURE "Fiorenzo Squarciafichi"
   PER IL PONENTE E LA SUA SICUREZZA (WATER WATCH)" basato sul lavoro
   originario WaterWatch 1.1 reperibile all'indirizzo:
   http://www.rotaryimperia.it/?page=waterwatch

   *********************************************************************
   Autori:
   Giacomo Galletti - 5B ITIS ind. Informatica e TLC
   Andrea Ruo - 5B INF - 5B ITIS ind. Informatica e TLC

   Supervisione:
   Ing. Simone Zanella - Docente di TPSIT presso ITIS ind. Informatica e TLC

   La documentazione completa del progetto ITISWW è disponibile su:
   http://www.polotecnologicoimperiese.gov.it/

   Versione 1.2 14-03-2017
   - prototipo iniziale

   *********************************************************************
   GNU General Public License

   This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    *********************************************************************
*/
//************************************
//dati centralina

char WW_id[] = "1234"; //ID dispositivo
String zona = "Demo"; //zona di monitoraggio dispositivo, per gli spazi usare %20
#define ALERT_SONAR 30 //soglia alert sonar in cm
#define WARNING_SONAR 100 //soglia warning sonar in cm
#define ALERT_BATTERIA 3.70 //soglia di avviso batteria scarica, da esprimere in volt:
/*    
    * 100% = 4,20[v]
    * 90% = 4.15[v]
    * 80% = 4.10[v]
    * 70% = 4.05 [v]
    * 60% = 4.00 [v]
    * 50% = 3.95 [v]
    * 40% = 3.90 [v]
    * 30% = 3.85 [v]
    * 20% = 3.80 [v]
    * 10% = 3.70 [v]
*/
#define LETTURA_ALLERTA 15000 //temporizzazione letture in condizioni di allerta (millisecondi)
#define LETTURA_STANDBY 30000 //temporizzazione letture in condizioni normali (millisecondi)

//************************************
//abilita/disabilita shield e sensori

/*
   I seguenti piedini non si possono usare: 0,1,2,3,7,10
*/

#define USE_YUN_SHIELD
//#define USE_GSM_SHIELD

//batteria
#define pin_batteria A10


//temperatura e umidità
#include "dht11.h"
#define pin_temp_umi 37
dht11 DHT;

//sensore pioggia
#define pin_pioggia A9

//ultrasuoni
#define sonar_signal A8

//pluviometro
#define pin_pluvi 18
#define interrupt_pluvi  digitalPinToInterrupt(pin_pluvi)
#define mm_impulso 0.142

//Thingspeak
#define myChannelNumber 225666
#define myWriteAPIKey "ZWE3S8QIGMD6UGNU"
String APIkey = String(myWriteAPIKey);

#if defined(USE_GSM_SHIELD)

  char server_dati[] = "api.thingspeak.com";
  int port = 80; //porta
  
#else if defined(USE_YUN_SHIELD)

  #include <ThingSpeak.h>
  
#endif

//IFTTT
#define KEY_IFTTT "p3e8VjMdl4sWyrWyEjH56vZDry3fRH01mKGZiNp9cUX"
String key = String(KEY_IFTTT);

#if defined(USE_GSM_SHIELD)

  char server_social[] = "maker.ifttt.com"; //URL IFTTT maker channel
  
#endif

/*
 * Messaggi per le trigger IFTTT:
 * 
 * SETUP
 * BATTERY
 * ALLERTA0
 * ALLERTA1
 * ALLERTA2
 * ALLERTA3
 * 
 */

//************************************
//segnalazione e sirena
#define frequenza_sirena 2000 //frequenza del suono in heartz
#define durata_suono 2000000 //durata in microsecondi del suono
#define periodo_sirena ((1000000/frequenza_sirena)) //calcola il periodo in microsecondi
#define pin_sirena 35
#define pin_LEDgiallo 33
#define pin_LEDrosso 31

//************************************
//shield

#if defined(USE_GSM_SHIELD)

  #include <GSM.h>
  #define PINNUMBER "" //PIN della SIM
  
  // APN data
  #define GPRS_APN       "wap.tim.it" // immettere l'APN GPRS
  #define GPRS_LOGIN     ""    // eventuale login
  #define GPRS_PASSWORD  "" // eventuale password
  
  // initialize the library instance
  GSMClient client;
  GPRS gprs;
  GSM gsmAccess;
  GSM_SMS sms;

  char senderNumber[20]= "";
  char txtMsg[160]="";
  
  boolean notConnected = true;

  int falliteGSM = 0;
  
  int lengthOfString;
  char *char_dati;
  void HTTPrequestGPRS(char server[], char path[]); //funzione per inviare le web request

#else if defined(USE_YUN_SHIELD)

  #include <Bridge.h>
  #include "YunClient.h"
  #include <HttpClient.h>
  
  BridgeClient Bclient;
  HttpClient Hclient;
  
#endif

//************************************
//variabili di sistema
int stato_allerta; //indica se ci sono allerta in corso: 0 = no, 1 = pioggia, 2 = sonar warning, 3 = sonar alert

//sensori
volatile float pluvi;
float sonar;
int temp;
int pioggia; //<=600 piove >600 non piove
int umi;
float battery;

//variabili temporanee
volatile float tmp_pluvi ;
float tmp_sonar;
int tmp_temp;
int tmp_pioggia;
int tmp_umi;
float tmp_battery;

//************************************
//interazione con i sensori

float Read_sonar();
void Read_pluvi();
int Read_temp();
int Read_pioggia(); //<=600 piove >600 non piove
float Read_battery();
int Read_umi();

void leggisensori();

void LEDgiallo(bool stato);
void LEDrosso(bool stato);
void sirena();

//************************************
//aggiornamento social e Thingspeak

void WWsalva(); //aggiorna Thingspeak
void WWsocial(int stato);//aggiorna i social
void WWalert(); //invia allarme soglia sonar critica superata
void WWwarning(); //invia avviso attenzione soglia sonar superata
void WWpioggia(); //invia avviso allerta pioggia/neve in corso
void WWmail(int batteria, String richiesta); //spedisce una mail di avviso livello basso batteria
void WWwait(); //temporizzazione letture

//************************************
//funzioni di controllo

void check_allerta();
void check_batteria();

//************************************
//funzioni di riavvio software
void(* ResetSW)(void) = 0;

//************************************

void setup() {

  //sensore temperatura
  //pinMode(pin_temp_umi, INPUT); //non serve se si usa analogico
  pinMode(pin_pluvi, INPUT);
  pinMode(pin_temp_umi, INPUT);
  pinMode(pin_pioggia, INPUT);
  pinMode(pin_sirena, OUTPUT);
  pinMode(pin_LEDgiallo, OUTPUT);
  pinMode(pin_LEDrosso, OUTPUT);
  pinMode(pin_batteria, INPUT);
  attachInterrupt(interrupt_pluvi, Read_pluvi, FALLING);
  
  detachInterrupt(interrupt_pluvi);
  
  #if defined(USE_GSM_SHIELD)

    Serial.begin(9600);
    while (!Serial) {
      ; // wait for serial port to connect. Needed for native USB port only
    }

    Serial.println("Inizializzazione Arduino Web Client...");
    while (notConnected) {
      
      if ((gsmAccess.begin(PINNUMBER) == GSM_READY) &
          (gprs.attachGPRS(GPRS_APN, GPRS_LOGIN, GPRS_PASSWORD) == GPRS_READY))
      {
        notConnected = false;
        Serial.println("Connessione GPRS completato");
      } 
      else 
      {
        Serial.println("Connessione GPRS fallita"); 
        delay(1000);
      }
      
    }
  
  #else if defined(USE_YUN_SHIELD)
  
    Bridge.begin();
    ThingSpeak.begin(Bclient);
  
  #endif
  
  //invia mail per segnalare avvio sistema
  WWmail(tmp_battery, "SETUP");
  
  //lettura iniziale
  leggisensori();
  WWsalva();
  
  interrupts();

}

void loop() {

  Serial.println("inizio loop");
  //check batteria
  check_batteria();
  
  //lettura sensori
  leggisensori();
#if defined(USE_GSM_SHIELD)

  //controllo SMS
  //se presente una richiesta invio tutti i dati via SMS all'operatore
  //la richiesta deve contenere l'ID del dispositivo altrimenti viene ignorata

  detachInterrupt(interrupt_pluvi);
  
  char c;
  char txt_tmp[160];
  String message = "";
  while (sms.available()){
    Serial.println("Messagio ricevuto da:"); //debug
 
    // numero del mittente
    sms.remoteNumber(senderNumber, 20);
    Serial.println(senderNumber); //debug

    while(c=sms.read()){
      message += c;
    }

    //controllo se ho richiesto i dati di questa centralina, altrimenti ignora
    if (message.equals(WW_id))
    {
      
      Serial.println("Invio dati"); //debug
      // cancella dalla memoria della SIM
      sms.flush();
      sms.beginSMS(senderNumber);

      //creo il messaggio con i dati
      strcat (txtMsg,"ID: ");
      strcat (txtMsg,WW_id);
      
      strcat (txtMsg," son");
      strcpy (txt_tmp, ""); //pulisce array
      sprintf(txt_tmp, ": %d", tmp_sonar); //float
      strcat (txtMsg,txt_tmp);

      strcat (txtMsg," pio");
      sprintf(txt_tmp, " %d", tmp_pioggia); //int
      strcat (txtMsg,txt_tmp);
      
      strcat (txtMsg," plu");
      strcpy (txt_tmp, ""); //pulisce array
      sprintf(txt_tmp, ": %d", tmp_pluvi); //float
      strcat (txtMsg,txt_tmp);
      
      strcat (txtMsg," tem");
      strcpy (txt_tmp, ""); //pulisce array
      sprintf(txt_tmp, ": %d", tmp_temp); //int
      strcat (txtMsg,txt_tmp);
      
      strcat (txtMsg," umi");
      strcpy (txt_tmp, ""); //pulisce array
      sprintf(txt_tmp, ": %d", tmp_umi); //int
      strcat (txtMsg,txt_tmp);
      
      strcat (txtMsg," bat");
      strcpy (txt_tmp, ""); //pulisce array
      sprintf(txt_tmp, ": %d", tmp_battery); //float
      strcat (txtMsg,txt_tmp);
     
      sms.print(txtMsg);
      sms.endSMS(); 
     
      delay(2000);
      
      Serial.println("Messaggio Inviato Correttamente"); 
    }
    else {
      // cancella dalla memoria della SIM
      sms.flush();
      Serial.println("Richiesta SMS ignorata"); 
    }
     
  
    } 

    interrupts();
    
#else if defined(USE_YUN_SHIELD)
  
  //al momento non c'è codice specifico per lo YUN
   
#endif

  //check allerta
  check_allerta();

  WWsalva(); //salva i valori e aggiorna Thingspeak

  WWwait(); //temporizzazione
}

//************************************
//corpo delle funzioni

float Read_sonar()
{
  detachInterrupt(interrupt_pluvi);
  //centimetri
  float lettura;

  //invia un impulso al sonar utilizzando il pin come output
  pinMode(sonar_signal, OUTPUT);
  digitalWrite(sonar_signal, HIGH);
  delayMicroseconds(20);
  digitalWrite(sonar_signal, LOW);

  //utilizza il pin come input per leggere la distanza
  pinMode(sonar_signal, INPUT);
  lettura = (float)pulseIn(sonar_signal, HIGH) * 34 / 2000.0;
  
  interrupts();
  return lettura;
}

void Read_pluvi() {
  detachInterrupt(interrupt_pluvi);
  tmp_pluvi += mm_impulso;
  interrupts();
}

int Read_temp()
{
  detachInterrupt(interrupt_pluvi);
  //gradi centigradi
  DHT.read(pin_temp_umi);

  int lettura = DHT.temperature;
  interrupts();
  return lettura;
}

int Read_pioggia()
{
  int lettura = analogRead(pin_pioggia);

  return lettura;
}

int Read_umi()
{

  detachInterrupt(interrupt_pluvi);
  
  //percentuale umidità
  DHT.read(pin_temp_umi);

  int lettura = DHT.humidity;
  interrupts();
  return lettura;
}

float Read_battery()
{
  float lettura = 0;

  lettura = (analogRead(pin_batteria)*0.049);

  return lettura;
}

void leggisensori()
{
  tmp_sonar = Read_sonar();
  tmp_temp = Read_temp();
  tmp_umi = Read_umi();
  tmp_pioggia = Read_pioggia();
  //la batteria viene controllata nel check
  //pluviometro si aggiorna separatamente via interrupt

  if ((pioggia <= 600) && (tmp_pioggia >600)){
    tmp_pluvi = 0;
    pluvi = 0;
  }
}

//************************************
//aggiornamento social e Thingspeak

void WWsalva() //aggiorna Thingspeak
{
  //salvo valori temporanei raccolti
  pluvi = tmp_pluvi;
  sonar = tmp_sonar;
  temp = tmp_temp;
  pioggia = tmp_pioggia;
  umi = tmp_umi;
  battery = tmp_battery;

  #if defined(USE_GSM_SHIELD)

    String tsData = "/update?api_key=" + APIkey;
    tsData = tsData + "&field1=" + String(sonar) + "&field2=" + String(temp) + "&field3=" + String(umi);
    tsData = tsData + "&field4=" + String(pluvi) + "&field5=" + String(battery) + "&field6=" + String(pioggia);

    //conversione
    lengthOfString=tsData.length();
    char_dati = new char [lengthOfString + 1];
    strcpy (char_dati, tsData.c_str()); 
    
    Serial.println("Invio dati a Thingspeak");
    HTTPrequestGPRS(server_dati, char_dati);

  #else if defined(USE_YUN_SHIELD)

  detachInterrupt(interrupt_pluvi);
    //salviamo tutti i dati su Thingspeak
    ThingSpeak.setField(1, sonar); //sonar
    ThingSpeak.setField(2, temp); //temp
    ThingSpeak.setField(3, umi); //umi
    ThingSpeak.setField(4, pluvi); //pluvi
    ThingSpeak.setField(5, battery); //carica batteria
    ThingSpeak.setField(6, pioggia); //sensore pioggia
    //ThingSpeak.setField(7, TRUE); //campo non utilizzato
    //ThingSpeak.setField(8,TRUE); //campo non utilizzato
  
    //aggiornamento multiplo
    ThingSpeak.writeFields(myChannelNumber, myWriteAPIKey);

    interrupts();
  
  #endif

  //thingspeak accetta aggiornamenti ogni 15 secondi
  //delay(15000);
}

void LEDgiallo(bool stato) {
  if (stato) {
    digitalWrite(pin_LEDgiallo, HIGH);
  }
  else {
    digitalWrite(pin_LEDgiallo, LOW);
  }
}

void LEDrosso(bool stato) {
  if (stato) {
    digitalWrite(pin_LEDrosso, HIGH);
  }
  else {
    digitalWrite(pin_LEDrosso, LOW);
  }
}

void sirena() {

  for (long i = 0; i < durata_suono; i += periodo_sirena)
  {
    digitalWrite(pin_sirena, HIGH);
    delayMicroseconds(periodo_sirena / 2);
    digitalWrite(pin_sirena, LOW);
    delayMicroseconds(periodo_sirena / 2);
  }

}

void WWmail(int batteria, String richiesta) //invia mail per livello batteria basso
{

  detachInterrupt(interrupt_pluvi);
  
  String codice;

  #if defined(USE_GSM_SHIELD)

  Serial.println("invio dati a IFTTT");
    codice = "https://maker.ifttt.com/trigger/BATTERY/with/key/" + key + "?value1=" + zona ;
    
    //conversione
    lengthOfString=codice.length();
    char_dati = new char [lengthOfString + 1];
    strcpy (char_dati, codice.c_str()); 

    HTTPrequestGPRS(server_social, char_dati);

  #else if defined(USE_YUN_SHIELD)
    detachInterrupt(interrupt_pluvi);
    codice = "https://maker.ifttt.com/trigger/" + richiesta + "/with/key/" + key + "?value1=" + zona + "&amp;value2=" + batteria;
    Hclient.get(codice);
    interrupts();
  #endif


}

void WWwait()
{
  //temporizzazione letture 
  
  if (stato_allerta > 0) //se >0 c'è almeno un evento piovoso in corso
  {
    delay(LETTURA_ALLERTA);
  }
  else
  {
    delay(LETTURA_STANDBY);
  }
}

void WWsocial(int stato = -1) //aggiorna i social
{
  String codice;

  switch (stato)
  {
    //lettura normale
    case -1:
      //non aggiorniamo niente
    break;

    //cessato allerta
    case 0:

      #if defined(USE_GSM_SHIELD)
        Serial.println("invio dati a IFTTT");
        codice = "https://maker.ifttt.com/trigger/ALLERTA0/with/key/" + key + "?value1=" + zona ;

        //conversione
        lengthOfString=codice.length();
        char_dati = new char [lengthOfString + 1];
        strcpy (char_dati, codice.c_str());
        
        HTTPrequestGPRS(server_social, char_dati);
        
      #else if defined(USE_YUN_SHIELD)

        detachInterrupt(interrupt_pluvi);
        codice = "https://maker.ifttt.com/trigger/ALLERTA0/with/key/" + key + "?value1=" + zona;
        Hclient.get(codice);
        interrupts();
        
      #endif

    break;

    //allerta pioggia
    case 1:
      
      #if defined(USE_GSM_SHIELD)
        Serial.println("invio dati a IFTTT");
        codice = "https://maker.ifttt.com/trigger/ALLERTA1/with/key/" + key + "?value1=" + zona ;

        //conversione
        lengthOfString=codice.length();
        char_dati = new char [lengthOfString + 1];
        strcpy (char_dati, codice.c_str());
        
        HTTPrequestGPRS(server_social, char_dati);
      
      #else if defined(USE_YUN_SHIELD)

        detachInterrupt(interrupt_pluvi);
        codice = "https://maker.ifttt.com/trigger/ALLERTA1/with/key/" + key + "?value1=" + zona;
        Hclient.get(codice);
        interrupts();
      #endif
      
    break;

    //warning sonar
    case 2:

      #if defined(USE_GSM_SHIELD)
        Serial.println("invio dati a IFTTT");
        codice = "https://maker.ifttt.com/trigger/ALLERTA2/with/key/" + key + "?value1=" + zona ;
        
        //conversione
        lengthOfString=codice.length();
        char_dati = new char [lengthOfString + 1];
        strcpy (char_dati, codice.c_str());
        
        HTTPrequestGPRS(server_social, char_dati);
      
      #else if defined(USE_YUN_SHIELD)

        detachInterrupt(interrupt_pluvi);
        codice = "https://maker.ifttt.com/trigger/ALLERTA2/with/key/" + key + "?value1=" + zona;
        Hclient.get(codice);
        interrupts();
        
      #endif
      
    break;

    //allerta sonar
    case 3:

      #if defined(USE_GSM_SHIELD)
        Serial.println("invio dati a IFTTT");
        codice = "https://maker.ifttt.com/trigger/ALLERTA3/with/key/" + key + "?value1=" + zona ;

        //conversione
        lengthOfString=codice.length();
        char_dati = new char [lengthOfString + 1];
        strcpy (char_dati, codice.c_str());
        
        HTTPrequestGPRS(server_social, char_dati);
      
      #else if defined(USE_YUN_SHIELD)

        detachInterrupt(interrupt_pluvi);
        codice = "https://maker.ifttt.com/trigger/ALLERTA3/with/key/" + key + "?value1=" + zona;
        Hclient.get(codice);
        interrupts();
      
      #endif
      
    break;
    
  }
  
}

void WWalert() //invia allarme soglia sonar critica superata //STATO 3
{

  //attiva led rosso fisso
  LEDgiallo(false);
  LEDrosso(true);

  //attiva sirena
  sirena();

  switch (stato_allerta)
  {
    case 0:
    case 1:
    case 2:
      //aumenta stato allerta
      WWsocial(3);
    break;

    case 3:
      //non fare niente, eravamo già in allerta
    break;
  }

}

void WWwarning() //invia avviso attenzione soglia sonar superata //STATO 2
{
  //attiva led giallo fisso
  LEDgiallo(true);
  LEDrosso(false);

  switch (stato_allerta)
  {
    case 0:
    case 1:
    case 3:
      //aumenta o cambia stato allerta
      WWsocial(2);
    break;

    case 2:
      //non fare niente, eravamo già in attenzione
    break;
  }

}

void WWpioggia() //invia avviso allerta pioggia/neve in corso //STATO 1
{
  LEDgiallo(false);
  LEDrosso(false);
  
  switch (stato_allerta)
  {
    case 0:
    case 2:
    case 3:
      //aumenta o cambia stato allerta
      WWsocial(1);
    break;

    case 1:
      //non fare niente, stava già piovendo
    break;
  }
}

//************************************
//funzioni di check
void check_batteria()
{
  tmp_battery = Read_battery();

  if (tmp_battery <= ALERT_BATTERIA)
  {
    WWmail(tmp_battery, "BATTERY");
  } 
}

void check_allerta()
{
  //controllo condizioni di allerta
  
  if (sonar <= ALERT_SONAR) //controllo il sonar
  {
    WWalert();
    
    //variabile globale
    stato_allerta = 3;
  }
  else if (sonar > ALERT_SONAR && sonar <= WARNING_SONAR) //controllo il sonar
  {
    WWwarning();
    
    //variabile globale
    stato_allerta = 2;
  }
  else if (pioggia <= 600) //controllo il sensore di pioggia
  {
    WWpioggia();
    
    //variabile globale
    stato_allerta = 1;
  }
  else
  {
    //spegni tutto
    LEDgiallo(false);
    LEDrosso(false);

    if (stato_allerta)
    {
      WWsocial(0); //se eravamo in stato di allerta, comunico la fine
    }
    
    stato_allerta = 0;
  }

  return stato_allerta;
}

#if defined(USE_GSM_SHIELD)

  void HTTPrequestGPRS(char server[], char path[])
  {

    detachInterrupt(interrupt_pluvi);
    Serial.println("Invio web request...");

    if (client.connect(server, 80)) {
      Serial.println("connected");
      client.print("GET ");
      client.print(path);
      client.println(" HTTP/1.1");
      client.print("Host: ");
      client.println(server);
      client.println("Connection: close");
      client.println();
      client.stop();

      falliteGSM = 0; //resetta il contatore connessioni fallite
    } 
    else {
      Serial.println("Connessione fallita nell'invio della web request");

      falliteGSM++; //incremento il contatore connessioni fallite
      
      if (falliteGSM > 3) {

        notConnected = true;
        
        Serial.println("Troppe connessioni fallite, riavvio il dispositivo...");
        interrupts();
        delay (1000);
        ResetSW();
      }
  }
  interrupts();
  }
#endif

